#!/usr/bin/env python3

from ansible.plugins.action import ActionBase

import os
import json
import functools


@functools.total_ordering
class MountPoint(object):
    def __init__(self, spec, device=None, partnum=None,
                 vg_name=None, lv_name=None):

        if not device and not (vg_name and lv_name):
            raise Exception("Either 'device' "
                            "or ('vg_name' and 'lv_name') "
                            "are required")

        self._spec = spec
        self.device = device
        self.partnum = partnum
        self.vg_name = vg_name
        self.lv_name = lv_name
        self.block_device = None

        self.path = self._spec.get('mount', {}).get('point')
        if self.path:
            self.path = os.path.normpath(self.path)

        self.opts = self._spec.get('mount', {}).get('opts')

    def device_path(self):
        if self.device:
            return self.device

        if self.vg_name and self.lv_name:
            return "/dev/mapper/{}-{}".format(self.vg_name.replace('-', '--'),
                                              self.lv_name.replace('-', '--'))

    def __eq__(self, other):
        return self.path == other.path

    def __lt__(self, other):
        return self.path < other.path

    def find_block_device(self, block_devices):
        block_device = self._find_block_device(block_devices)
        if block_device:
            self.block_device = dict(filter(lambda x: x[0] != 'children',
                                            block_device.items()))
        else:
            self.block_device = {}

    def _find_block_device(self, block_devices):
        device_path = self.device_path()
        for device in block_devices:
            if device['name'] == device_path:
                if self.partnum:
                    try:
                        return device['children'][self.partnum - 1]
                    except IndexError:
                        return None
                else:
                    return device

            if 'children' in device:
                block_device = self._find_block_device(device['children'])
                if block_device:
                    return block_device

    def to_dict(self):
        data = {
            'path': self.path,
            'blockdevice': self.block_device,
            'uuid': self.block_device.get('uuid'),
            'fs': self._spec['filesystem']['type'],
        }
        # NOTE: Add 'opts' key only if self.opts is a non-empty string.
        #       This is to use Ansible's 'default' filter later. It only
        #       works if key is not exist.
        if self.opts:
            data['opts'] = self.opts
        return data


class ActionModule(ActionBase):
    def __init__(self, *args, **kwargs):
        self._tmp = None
        self._task_vars = None
        super(ActionModule, self).__init__(*args, **kwargs)

    def _get_block_devices(self):
        module_args = {
            'confirm_uuid': True,
        }

        result = self._execute_module(module_name='lsblk',
                                      module_args=module_args,
                                      task_vars=self._task_vars,
                                      tmp=self._tmp)

        return result.get('blockdevices', [])

    def _get_md_device(self, name):
        module_args = {
            'name': name,
            'state': 'info',
        }

        result = self._execute_module(module_name='mdadm',
                                      module_args=module_args,
                                      task_vars=self._task_vars,
                                      tmp=self._tmp)

        return result['real_dev']

    def _mountpoints(self, target_storage):
        mountpoints = []
        for device_spec in target_storage:
            device_type = device_spec.get('type')
            if device_type == 'physical':
                device_name = device_spec.get(
                    device_spec.get('work_by', 'by_name'))
                for part_spec in device_spec.get('partition_schema', []):
                    if part_spec.get('mount', {}).get('point'):
                        mountpoints.append(
                            MountPoint(part_spec,
                                       device=device_name,
                                       partnum=part_spec['num']))
            elif device_type == 'lvm':
                for vg_spec in device_spec.get('lvm_groups', []):
                    for lv_spec in vg_spec.get('lvnames', []):
                        if lv_spec.get('mount', {}).get('point'):
                            mountpoints.append(
                                MountPoint(lv_spec,
                                           vg_name=vg_spec['vgname'],
                                           lv_name=lv_spec['lvname']))
            elif device_type == 'md':
                device_name = self._get_md_device(device_spec['name'])
                if device_spec.get('mount', {}).get('point'):
                    mountpoints.append(MountPoint(device_spec, device_name))
            else:
                raise Exception("Device type '{}' is not supported"
                                .format(device_type))

        # FIXME: refactor to PATH logic-sort check
        block_devices = self._get_block_devices()
        for mountpoint in mountpoints:
            mountpoint.find_block_device(block_devices)

        return [x.to_dict() for x in sorted(mountpoints)]

    def run(self, tmp=None, task_vars=None):
        self._supports_check_mode = False
        self._supports_async = False
        self._tmp = tmp
        self._task_vars = task_vars

        result = super(ActionModule, self).run(tmp, task_vars)

        result['ordered_mounts'] = self._mountpoints(
            self._task.args['target_storage'])

        return result
