#!/usr/bin/python3
"""
'slice_block_devices' filter transforms tree-like structure of block devices
into 'slices' where each slice is a list of block devices at the same level
of hierarchy.

Consider the following tree-like structure:

    blockdevices:
      - children:
        - kname: /dev/sda1
          name: /dev/sda1
          pkname: /dev/sda
          size: '1048576'
          type: part
        - children:
          - kname: /dev/md127
            name: /dev/md127
            pkname: /dev/sda2
            size: '299824578560'
            type: raid1
          kname: /dev/sda2
          name: /dev/sda2
          pkname: /dev/sda
          size: '299959844864'
          type: part
        kname: /dev/sda
        name: /dev/sda
        pkname: null
        size: '299966445568'
        type: disk

It will be transformed to these slices:

    -
      - kname: /dev/sda
        name: /dev/sda
        pkname: null
        size: '299966445568'
        type: disk
    -
      - kname: /dev/sda1
        name: /dev/sda1
        pkname: /dev/sda
        size: '1048576'
        type: part
      - kname: /dev/sda2
        name: /dev/sda2
        pkname: /dev/sda
        size: '299959844864'
        type: part
    -
      - kname: /dev/md127
        name: /dev/md127
        pkname: /dev/sda2
        size: '299824578560'
        type: raid1
"""


def _slice_block_devices(block_devices, layers=None, level=0):
    if layers is None:
        layers = {}

    for block_device in block_devices:
        # Copy all fields except 'children' to a new dict and
        # add it at the appropriate level.
        layers.setdefault(level, []).append(dict(
            (k, v) for k, v in block_device.items() if k != "children"
        ))
        # Process children devices recursively
        if "children" in block_device:
            _slice_block_devices(block_device["children"], layers, level+1)
    if level == 0:
        return [layers[x] for x in sorted(layers)]


def slice_block_devices(block_devices):
    return _slice_block_devices(block_devices)


class FilterModule(object):
    def filters(self):
        return {
            'slice_block_devices': slice_block_devices,
        }
