#!/usr/bin/env python3
# Copyright 2020 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ansible.module_utils.basic import AnsibleModule

import json


ANSIBLE_METADATA = {
    'metadata_version': '0.0.1',
    'status': ['preview'],
    'supported_by': 'dev@mirantis.com'
}

DOCUMENTATION = r'''
---
author:
 - Mirantis (dev@mirantis.com)
module: lvm
short_description: Configure LVM volumes
version_added: "0.0.1"
description:
  - 'This module allows configuring LVM volumes'
requirements:
  - ['lvm2']
'''


def run_command(cmd, fail_msg='', ok_exit_codes=(0, )):
    global module

    exit_code, stdout, stderr = module.run_command(cmd)
    if exit_code not in ok_exit_codes:
        module.fail_json(
            msg="{} cmd: {} exit_code: {} stdout: {} stderr: {}"
                .format(fail_msg, cmd, exit_code, stdout, stderr))

    return exit_code, stdout, stderr


def _lvm_report(comm, key, units=None, options=None, suffix=True):
    """
    Collect LVM report for one of the following LVM objects - lv, vg, pv.

    For example, 'lvs' command returns the following structure:

    $ lvs --reportformat json --units m
      --options lv_name,lv_size,vg_name,lv_uuid,lv_path

    {
      "report": [
        {
          "lv": [
            {
              "lv_name":"lvroot",
              "lv_size":"665600,00m",
              "vg_name":"sysvg",
              "lv_uuid":"0a86308e-dbd4-11ea-8fdf-cf392ceb4881",
              "lv_path":"/dev/sysvg/lvroot"
            },
            {
              "lv_name":"lvswap",
              "lv_size":"16384,00m",
              "vg_name":"sysvg",
              "lv_uuid":"0b10cf32-dbd4-11ea-b1db-cf2e8e2020d9",
              "lv_path":"/dev/sysvg/lvswap"
            }
          ]
        }
      ]
    }

    For this structure this method returns generator that will iterate
    over the following list (note that there is no single variable that
    contains this list):

    [
      {
        "lv_name":"lvroot",
        "lv_size":"665600,00m",
        "vg_name":"sysvg",
        "lv_uuid":"0a86308e-dbd4-11ea-8fdf-cf392ceb4881",
        "lv_path":"/dev/sysvg/lvroot"
      },
      {
        "lv_name":"lvswap",
        "lv_size":"16384,00m",
        "vg_name":"sysvg",
        "lv_uuid":"0b10cf32-dbd4-11ea-b1db-cf2e8e2020d9",
        "lv_path":"/dev/sysvg/lvswap"
      }
    ]

    :param comm: LVM command to collect data. Allowed commands are:
        * 'pvs'
        * 'vgs'
        * 'lvs'.
    :param key: Key to fetch data from LVM report. This key is usually
        1:1 mapped with 'comm' value:
            comm | key
            -----+------
            pvs  | pv
            vgs  | vg
            lvs  | lv
    :param units: units to format volume size
    :param options: fields that should be included in the report,
        comma-separated
    :return: generator that iterates over values stored under 'key'
        in the returned report structure.
    """
    global module

    if comm not in ['pvs', 'vgs', 'lvs']:
        module.fail_json(msg="Invalid LMV command '{}'".format(comm))

    cmd = [comm, '--reportformat', 'json']
    if not suffix:
        cmd.append('--nosuffix')
    if units:
        cmd.extend(['--units', units])
    if options:
        cmd.extend(['--options', options])

    _, stdout, _ = run_command(cmd, fail_msg=f'Failed to get {comm} report')

    for report in json.loads(stdout)['report']:
        for record in report.get(key, []):
            yield record


def _lv_data(options='lv_name,lv_size,vg_name,lv_uuid,lv_path'):
    return _lvm_report('lvs', 'lv', units='b', suffix=False, options=options)


def _vg_data(options='vg_name,vg_uuid,vg_size,vg_free'):
    return _lvm_report('vgs', 'vg', units='b', suffix=False, options=options)


def _pv_data(options='pv_name,vg_name,pv_size,dev_size,pv_uuid'):
    return _lvm_report('pvs', 'pv', units='b', suffix=False, options=options)


def _lvremove(lv_path=None, lv_uuid=None):
    global module

    if lv_path:
        # check if lv exists
        if not get_first_by_key_value(_lv_data(), 'lv_path', lv_path):
            module.fail_json(
                msg=f'Error while removing lv: lv {lv_path} not found')
        cmd = ['lvremove', '-ff', lv_path]
    elif lv_uuid:
        # check if lv exists
        if not get_first_by_key_value(_lv_data(), 'lv_uuid', lv_uuid):
            module.fail_json(
                msg=f'Error while removing lv: lv {lv_uuid} not found')
        cmd = ['lvremove', '-ff', '--select', f'UUID={lv_uuid}']
    else:
        module.fail_json(msg='Invocation error')

    run_command(cmd, fail_msg='lvremove failed!')


def _vgremove(vg_name=None, vg_uuid=None):
    global module

    if vg_name:
        if not get_first_by_key_value(_vg_data(), 'vg_name', vg_name):
            module.fail_json(
                msg=f'Error while removing vg: {vg_name} not found')
        cmd = ['vgremove', '-ff', vg_name]
    elif vg_uuid:
        if not get_first_by_key_value(_vg_data(), 'vg_uuid', vg_uuid):
            module.fail_json(
                msg=f'Error while removing vg:{vg_uuid} not found')
        cmd = ['vgremove', '-ff', '--select', f'UUID={vg_uuid}']
    else:
        module.fail_json(msg='Invocation error')

    run_command(cmd, fail_msg='vgremove failed!')


def _pvremove(pv_name):
    """
    pvremove - disk-path, since pvremove work only with path)
    :param kwargs:
    :return:
    """
    global module

    pv = get_first_by_key_value(_pv_data(), 'pv_name', pv_name)

    # check if pv exists
    if not pv:
        module.fail_json(
            msg=f'Error while removing pv: pv {pv_name} not found')

    # check if pv is attached to some vg
    if pv['vg_name']:
        module.fail_json(
            msg=f'Error while removing {pv}: pv belongs to vg {pv["vg_name"]}')

    cmd = ['pvremove', '-ff', '--yes', pv_name]
    run_command(cmd, fail_msg='vgremove failed!')


def lvremove(spec):
    global module

    if 'lv_name' in spec or 'lv_uuid' in spec:
        return _lvremove(lv_path=spec.get('lv_path'),
                         lv_uuid=spec.get('lv_uuid'))
    else:
        module.fail_json(
            msg="'lv' spec should countain either 'lv_name' or 'lv_uuid'")


def vgremove(spec):
    global module

    if 'vg_name' in spec or 'vg_uuid' in spec:
        return _vgremove(vg_name=spec.get('vg_name'),
                         vg_uuid=spec.get('vg_uuid'))
    else:
        module.fail_json(
            msg="'vg' spec should countain either 'vg_name' or 'vg_uuid'")


def pvremove(spec):
    global module

    if 'pv_name' in spec:
        return _pvremove(spec.get('pv_name'))
    else:
        module.fail_json(msg="'pv' spec should countain 'pv_name'")


def get_first_by_key_value(collection, key, value, default=None):
    return next((x for x in collection if x[key] == value), default)


def get_lvm_from_disk(pv_name=None, vg_name=None, lv_name=None):
    """
    Collect all resources UUID from phys device.
    TODO: those might not be valid, if one lv spanned across multiply disks
    In that case we might use lv\vg-reduce

    Optional pv_name / vg_name / lv_name allows to return only objects that
    are related to requested PV / VG / LV.

    :param pv_name: include VGs/LVs that depends on that PV
    :param vg_name: include LVs that depend on that VG and / or
                    PVs that are used by that VG
    :param lv_name: include PVs/VGs that used by that LV
    :return:
    lvs = List of LV UUID
    vgs = List of vgs UUID
    pvs = List of pv_name (disk-path, since pvremove work only with path)
    """
    global module

    pvs = []
    lvs = []
    vgs = []

    # Collect PVs that match requested PV / VG
    for pv in _pv_data():
        if pv_name and pv['pv_name'] != pv_name:
            continue

        if vg_name and pv['vg_name'] != vg_name:
            continue

        pvs.append(pv)

    # Collect VGs that belong to PVs
    if pvs:
        for vg in _vg_data():
            if vg_name and vg['vg_name'] != vg_name:
                continue

            # Test if VG belongs to any PV in the PVs list
            if any(filter(lambda x: x['vg_name'] == vg['vg_name'], pvs)):
                vgs.append(vg)

    # Collect LVs that belong to VGs
    if vgs:
        for lv in _lv_data():
            if vg_name and lv['vg_name'] != vg_name:
                continue

            if lv_name and lv['lv_name'] != lv_name:
                continue

            # Test if LV belongs to any VG in the VGs list
            if any(filter(lambda x: x['vg_name'] == lv['vg_name'], vgs)):
                lvs.append(lv)

    return {'pvs': pvs, 'vgs': vgs, 'lvs': lvs}


# def pvcreate(pvname, metadatasize=128, metadatacopies=2):
#     # check if pv already exists
#     global module
#
#     if get_first_by_key_value(_pv_data(), 'pv_name', pvname):
#         module.fail_json(
#             msg=f'Error while creating pv: pv {pvname} already exists')
#
#     run_command(['pvcreate',
#                  '--metadatacopies', str(metadatacopies),
#                  '--metadatasize', f'{metadatasize}m',
#                  pvname], fail_msg='pvcreate failed!')
#
#
# def lvcreate(vgname, lvname, size):
#     vg = get_first_by_key_value(vgdisplay(), 'name', vgname)
#
#     # check if vg exists
#     if not vg:
#         raise errors.VGNotFoundError(
#             'Error while creating vg: vg %s not found' % vgname)
#     # check if enough space is available
#     if vg['free'] < size:
#         raise errors.NotEnoughSpaceError(
#             'Error while creating lv: vg %s has only %s m of free space, '
#             'but at least %s m is needed' % (vgname, vg['free'], size))
#     # check if lv already exists
#     if next(
#             (x for x in lvdisplay() if
#              x['name'] == lvname and x['vg'] == vgname),
#             False
#     ):
#         raise errors.LVAlreadyExistsError(
#             'Error while creating lv: lv %s already exists' % lvname)
#     # NOTE(agordeev): by default, lvcreate is configured to wipe signature
#     # on allocated volume. '--yes' should be passed to avoid waiting for
#     # user's confirmation:
#     # "WARNING: <signature> signature detected on <device>. Wipe it? [y/n]"
#     # FIXME: the version of lvm2 shipped with Ubuntu 14.04 does not support
#     # --yes option. fuel-agent should properly decomission the storage
#     # (Ubuntu installer does that just fine).
#     stdout, stderr = utils.execute('lvcreate', '--help')
#     force_opt = '--yes' if '--yes' in stdout else ''
#     cmd = 'lvcreate {force_opt} -L {size}m -n {lvname} {vgname}'.format(
#         size=size, lvname=lvname, vgname=vgname, force_opt=force_opt)
#     utils.execute(*cmd.split(), check_exit_code=[0])
#
#
# def vgcreate(vgname, pvname, *args):
#     # check if vg already exists
#     if get_first_by_key_value(vgdisplay(), 'name', vgname, False):
#         raise errors.VGAlreadyExistsError(
#             'Error while creating vg: vg %s already exists' % vgname)
#     pvnames = [pvname] + list(args)
#     _vg_attach_validate(pvnames)
#     utils.execute('vgcreate', vgname, *pvnames, check_exit_code=[0])
#
#
# def vgextend(vgname, pvname, *args):
#     # check if vg exists
#     if not get_first_by_key_value(vgdisplay(), 'name', vgname, False):
#         raise errors.VGNotFoundError(
#             'Error while extending vg: vg %s not found' % vgname)
#     pvnames = [pvname] + list(args)
#     _vg_attach_validate(pvnames)
#     utils.execute('vgextend', vgname, *pvnames, check_exit_code=[0])
#
#
# def vgreduce(vgname, pvname, *args):
#     # check if vg exists
#     if not get_first_by_key_value(vgdisplay(), 'name', vgname, False):
#         raise errors.VGNotFoundError(
#             'Error while reducing vg: vg %s not found' % vgname)
#     pvnames = [pvname] + list(args)
#     # check if all necessary pv are attached to vg
#     if not set(pvnames).issubset(
#             set([pv['name'] for pv in pvdisplay() if pv['vg'] == vgname])):
#         raise errors.PVNotFoundError(
#             'Error while reducing vg: at least one of pv is '
#             'not attached to vg')
#     utils.execute('vgreduce', '-f', vgname, *pvnames, check_exit_code=[0])


def run_module():
    global module

    changed = False
    module = AnsibleModule(
        argument_spec=dict(
            pv=dict(type='dict'),
            pvs=dict(type='list'),
            vg=dict(type='dict'),
            vgs=dict(type='list'),
            lv=dict(type='dict'),
            lvs=dict(type='list'),
            match=dict(type='dict', default={}),
            state=dict(type='str', default='info',
                       choices=['absent', 'info', 'present'])
        ),
        supports_check_mode=False,
    )
    module.run_command_environ_update = {
        'LANG': 'C',
        'LC_ALL': 'C',
        'LC_MESSAGES': 'C',
        'LC_CTYPE': 'C'
    }

    lv_spec = module.params['lv']
    lv_spec_list = module.params['lvs']
    vg_spec = module.params['vg']
    vg_spec_list = module.params['vgs']
    pv_spec = module.params['pv']
    pv_spec_list = module.params['pvs']
    match = module.params['match']
    state = module.params['state']

    if not lv_spec_list and lv_spec:
        lv_spec_list.append(lv_spec)

    if not vg_spec_list and vg_spec:
        vg_spec_list.append(vg_spec)

    if not pv_spec_list and pv_spec:
        pv_spec_list.append(pv_spec)

    lvm_data = {}
    if state == 'present':
        module.fail_json(msg="Not supported")
    elif state == 'absent':
        for lv_spec in lv_spec_list:
            lvremove(lv_spec)
            changed = True
        for vg_spec in vg_spec_list:
            vgremove(vg_spec)
            changed = True
        for pv_spec in pv_spec_list:
            pvremove(pv_spec)
            changed = True
    elif state == 'info':
        lvm_data = get_lvm_from_disk(pv_name=match.get('pv'),
                                     vg_name=match.get('vg'),
                                     lv_name=match.get('lv'))

    module.exit_json(
        changed=changed,
        lvs=lvm_data.get('lvs', []),
        pvs=lvm_data.get('pvs', []),
        vgs=lvm_data.get('vgs', []),
    )


if __name__ == '__main__':
    run_module()
